//BSD License - please cite paper for good karma :-)
/*
Copyright 2011 David Koes. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of David Koes.
*/

/*
 * data.h
 *
 *  Created on: Apr 25, 2011
 *      Author: dkoes
 *
 *  Contains classes for storing data.
 */


#ifndef DATA_H_
#define DATA_H_

#include <iostream>
#include <boost/unordered_map.hpp>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <algorithm>
#include <cmath>
#include <cstdio>

#include "CommandLine2/CommandLine.h"
extern cl::opt<bool> Verbose;
extern cl::opt<unsigned> Debug;

using namespace std;
using namespace boost;

/*
 * A rule for splitting data
 */
struct Split
{
	unsigned index; //attribute
	double split; //value
	bool greater; //if true, select > split

	Split(): index(0), split(0), greater(false)
	{

	}

	Split(unsigned i, double sp, bool gt): index(i), split(sp), greater(gt)
	{

	}
};

//function object for splitting a dataview on a conjunctive rule
class AndSplitter
{
	const vector<Split>& rule;
public:
	AndSplitter(const vector<Split>& r): rule(r)
	{

	}

	//return true if matches the rule
	bool operator()(double *vals) const
	{
		bool ret = true;
		for(unsigned i = 0, n = rule.size(); i < n; i++)
		{
			unsigned index = rule[i].index;
			bool r = vals[index] < rule[i].split;
			if(rule[i].greater)
				r = !r;
			ret &= r;
		}
		return ret;
	}
};

//stores statics for applying a rule
struct Stats
{
	double tp;
	double fp;
	double tn;
	double fn;

	Stats(): tp(0), fp(0), tn(0), fn(0) {}
	Stats(unsigned TP, unsigned FP, unsigned TN, unsigned FN): tp(TP), fp(FP), tn(TN), fn(FN)
	{

	}

	double accuracy() const
	{
		return (tp+tn)/(tp+tn+fp+fn);
	}

	double precision() const
	{
		if(tp+fp == 0) return 0;
		return tp/(tp+fp);
	}

	double recall() const
	{
		if(tp+fn == 0) return 0;
		return tp/(tp+fn);
	}

	double specificity() const
	{
		if(tn+fp == 0) return 0;
		return tn/(tn+fp);
	}

	double F1() const
	{
		double p = precision();
		double r = recall();
		if(p+r == 0) return 0;
		return (2*p*r)/(p+r);
	}

};

/* Information on each column of data (ie, name, maybe eventually type) */
class DataSchema
{
	vector<string> names;
	unordered_map<string, unsigned> nameIndex;
	vector<bool> reverse; //if true, use opposite direction than calculated for this attribute
public:
	DataSchema(const string& header)
	{
		//should just be names, at some point may contain type information
		//just tokenize
		stringstream str(header);
		string name;
		str >> name; //ignore label col
		while (str >> name)
		{
			nameIndex[name] = names.size();
			names.push_back(name);
			reverse.push_back(false);
		}
	}

	bool isValid() const
	{
		return names.size() > 0; //first name is label column
	}

	//index of attribute, -1 if invalid
	int index(const string& name) const
	{
		unordered_map<string, unsigned>::const_iterator itr = nameIndex.find(
				name);
		if (itr == nameIndex.end())
			return -1;
		return itr->second;
	}

	const string& name(unsigned index) const
	{
		return names[index];
	}

	void setReverse(const string& name)
	{
		int i = index(name);
		if(i >= 0)
		{
			reverse[i] = true;
		}
	}

	bool getReverse(unsigned i)
	{
		if(i < reverse.size())
		{
			return reverse[i];
		}
		return false;
	}
	//number of attributes (ignoring label)
	unsigned size() const
	{
		return names.size();
	}
};

//actual data; every data item is a double
//binary classification is assumed and data is split based on
//first column
class Data
{
	vector<double *> truedata; //store rows as pointers for faster sorting
	vector<double *> falsedata;
	unsigned ncols;

public:

	virtual ~Data()
	{
		for (unsigned i = 0, n = truedata.size(); i < n; i++)
		{
			free(truedata[i]);
		}
		for (unsigned i = 0, n = falsedata.size(); i < n; i++)
		{
			free(falsedata[i]);
		}
	}

	Data() :
		ncols(0)
	{
	}

	Data(const Data& rhs)
	{
		//copy constructor, do a deep copy
		ncols = rhs.ncols;
		truedata.reserve(rhs.truedata.size());
		for (unsigned i = 0, n = rhs.truedata.size(); i < n; i++)
		{
			double *tdata = (double*) malloc(sizeof(double) * ncols);
			memcpy(tdata, rhs.truedata[i], sizeof(double) * ncols);
			truedata.push_back(tdata);
		}

		falsedata.reserve(rhs.falsedata.size());
		for (unsigned i = 0, n = rhs.falsedata.size(); i < n; i++)
		{
			double *fdata = (double*) malloc(sizeof(double) * ncols);
			memcpy(fdata, rhs.falsedata[i], sizeof(double) * ncols);
			falsedata.push_back(fdata);
		}

	}

	//read in file a line at a time, data must be whitespace separated
	Data(const DataSchema& schema, istream& in) :
		ncols(schema.size())
	{
		string line;

		//allocate a block per a line
		while (getline(in, line))
		{
			stringstream str(line);
			//resize buffer
			double *data = (double*) malloc(sizeof(double) * ncols);
			//absorb non numerical data
			string datum;
			//label
			str >> datum;
			double tf = atof(datum.c_str());
			for (unsigned i = 0; i < ncols; i++)
			{
				str >> datum;
				double val = atof(datum.c_str());
				data[i] = val;
			}
			if (tf == 0)
				falsedata.push_back(data);
			else
				truedata.push_back(data);
		}
	}

	//number of data samples
	unsigned size() const
	{
		return truedata.size() + falsedata.size();
	}

	friend class DataView;
};

//a view of data - does not own memory, can freely sort and filter rows
class DataView
{
	unsigned ncols;
	vector<double *> truedata; //store rows as pointers for faster sorting
	vector<double *> falsedata;

	struct Sorter
	{
		unsigned index;
		Sorter(unsigned i) :
			index(i)
		{
		}

		bool operator()(const double *left, const double *right) const
		{
			return left[index] < right[index];
		}
	};
public:

	virtual ~DataView()
	{
		//does not own memory
	}

	DataView() :
		ncols(0)
	{
	}

	//copy constructor, just a pointer copy since we don't own memory
	DataView(const DataView& rhs): ncols(rhs.ncols), truedata(rhs.truedata), falsedata(rhs.falsedata)
	{
	}

	//pointer copy out of Data object
	DataView(const Data& par): ncols(par.ncols), truedata(par.truedata), falsedata(par.falsedata)
	{

	}

	void invert()
	{
		swap(truedata, falsedata);
	}

	//number of data samples
	unsigned size() const
	{
		return truedata.size() + falsedata.size();
	}

	unsigned truesize() const
	{
		return truedata.size();
	}

	unsigned falsesize() const
	{
		return falsedata.size();
	}

	//Shannon information entropy
	double entropy() const
	{
		if(size() == 0)
			return 0;
		double trueratio = truesize()/(double)size();
		double falseratio = falsesize()/(double)size();

		if(falseratio == 0)
			return 0;
		if(trueratio == 0)
			return 0;

		double H = -trueratio*log2(trueratio) - falseratio*log2(falseratio);

		return H;
	}

	//sort ascending on index
	void sort(unsigned index)
	{
		Sorter sort(index);
		std::sort(truedata.begin(), truedata.end(), sort);
		std::sort(falsedata.begin(), falsedata.end(), sort);
	}

	//reduce number of items using selector (return true to keep)
	//gets passed all column values and truth value
	template<class Selector>
	void filter(Selector& select)
	{
		vector<double*> newtrue; newtrue.reserve(truedata.size());
		for(unsigned i = 0, n = truedata.size(); i < n; i++)
		{
			if(select(truedata[i], true))
				newtrue.push_back(truedata[i]);
		}
		truedata.swap(newtrue);

		vector<double*> newfalse; newfalse.reserve(falsedata.size());
		for(unsigned i = 0, n = falsedata.size(); i < n; i++)
		{
			if(select(falsedata[i], false))
				newtrue.push_back(falsedata[i]);
		}
		falsedata.swap(newfalse);
	}

	//uses select to separate this data view into two disjoint views
	//overwrites dataviews
	template<class Selector>
	void splitter(Selector& select, DataView& selected, DataView& unselected) const
	{
		selected.ncols = ncols;
		selected.truedata.clear(); selected.truedata.reserve(truesize());
		selected.falsedata.clear(); selected.falsedata.reserve(falsesize());
		unselected.ncols = ncols;
		unselected.truedata.clear(); unselected.truedata.reserve(truesize());
		unselected.falsedata.clear(); unselected.falsedata.reserve(falsesize());

		for(unsigned i = 0, n = truedata.size(); i < n; i++)
		{
			if(select(truedata[i]))
				selected.truedata.push_back(truedata[i]);
			else
				unselected.truedata.push_back(truedata[i]);
		}

		for(unsigned i = 0, n = falsedata.size(); i < n; i++)
		{
			if(select(falsedata[i]))
				selected.falsedata.push_back(falsedata[i]);
			else
				unselected.falsedata.push_back(falsedata[i]);
		}
	}

	//iterators
	class iterator
	{
		int index;
		unsigned currtrue; //position in true array
		unsigned currfalse; //position in false array
		double *currarray;
		const DataView *parent;

		friend class DataView;
		iterator(const DataView* p, int i) :
			index(i), currtrue(0), currfalse(0), currarray(NULL), parent(p)
		{
			if (index < 0)
				currarray = parent->truedata[0];
			else
				currarray = parent->truedata[0][index]
						<= parent->falsedata[0][index] ? parent->truedata[0]
						: parent->falsedata[0];
		}

	public:
		iterator() :
			index(-1), currtrue(0), currfalse(0), currarray(NULL), parent(NULL)
		{
		}

		bool operator==(const iterator& rhs) const
		{
			if (currarray == NULL && rhs.currarray == NULL)
				return true; //both invalid
			return currarray == rhs.currarray && index == rhs.index && currtrue
					== rhs.currtrue && currfalse == rhs.currfalse && parent
					== rhs.parent;
		}

		operator bool() const
		{
			return currarray != NULL && parent != NULL && (currtrue
					< parent->truesize() || currfalse < parent->falsesize());
		}

		//is this instance classified as true?
		bool isTrue() const
		{
			return (currtrue < parent->truesize() && currarray
					== parent->truedata[currtrue]);
		}

		//the value at the iterating index
		double value() const
		{
			if(index < 0)
				return 0;
			if(currarray == NULL)
				return 0;
			return currarray[index];
		}

		iterator& operator++()
		{
			if (currtrue >= parent->truesize())
			{
				if (currfalse >= parent->falsesize())
					return *this;
				currfalse++;
				if (currfalse == parent->falsesize())
					currarray = NULL;
				else
					currarray = parent->falsedata[currfalse];
				return *this;
			}
			else if (currfalse >= parent->falsesize())
			{
				currtrue++;
				if (currtrue == parent->truesize())
					currarray = NULL;
				else
					currarray = parent->truedata[currtrue];
				return *this;
			}
			else if (index < 0) //ignore values of attributes
			{
				currtrue++;
				if (currtrue == parent->truesize())
					currarray = parent->falsedata[currfalse];
				else
					currarray = parent->truedata[currtrue];
				return *this;
			}
			else
			{
				//choose next smallest attribute
				//if current value is from true data
				if (currarray == parent->truedata[currtrue])
				{
					//compare next true to current false;
					currtrue++;
					if (currtrue == parent->truesize())
					{
						currarray = parent->falsedata[currfalse];
					}
					else
					{
						currarray
								= parent->truedata[currtrue][index]
										<= parent->falsedata[currfalse][index] ? parent->truedata[currtrue]
										: parent->falsedata[currfalse];
					}
				}
				else if (currarray == parent->falsedata[currfalse])
				{
					//from false
					currfalse++;
					if (currfalse == parent->falsesize())
					{
						currarray = parent->truedata[currtrue];
					}
					else
					{
						currarray
								= parent->truedata[currtrue][index]
										<= parent->falsedata[currfalse][index] ? parent->truedata[currtrue]
										: parent->falsedata[currfalse];

					}
				}
				else
					abort();
			}
			return *this;
		}

		//skip to the next distinct value
		void nextValue()
		{
			if(index < 0)
				++(*this);
			else
			{
				//TODO: make this more efficient
				double val = currarray[index];
				while(currarray != NULL && currarray[index] == val)
				{
					++(*this);
				}
			}
		}

		const double* operator*() const
		{
			return currarray;
		}

		//number of values (not including current) that have been iterated past
		unsigned seenCnt() const
		{
			return currtrue+currfalse;
		}

		//number of true values (not including current) that have been iterated past
		unsigned seenTrueCnt() const
		{
			return currtrue;
		}

		//number of false values iterated past
		unsigned seenFalseCnt() const
		{
			return currfalse;
		}
	};

	//iterate over true than false
	iterator begin() const
	{
		iterator itr(this, -1);
		return itr;
	}

	iterator end() const
	{
		iterator itr(this, -1);
		itr.currarray = NULL;
		return itr;
	}

	//iterate over attributes in sorted order (undefined results if not sorted on attrib i)
	iterator begin(unsigned i) const
	{
		return iterator(this, i);
	}

	unsigned columns() const
	{
		return ncols;
	}

	//apply a rule and compute accuracy statistics
	Stats applyRule(const vector<Split>& rule, bool biasfalse)
	{
		AndSplitter splitter(rule);

		unsigned tp = 0, fp = 0, tn = 0, fn = 0;
		for(unsigned i = 0, n = truedata.size(); i < n; i++)
		{
			if(splitter(truedata[i]))
				tp++;
			else
				fn++;
		}
		for(unsigned i = 0, n = falsedata.size(); i < n; i++)
		{
			if(splitter(falsedata[i]))
				fp++;
			else
				tn++;
		}

		if(biasfalse)
		{
			swap(tp,fp);
			swap(tn,fn);
		}
//printf("tp %d fp %d tn %d fn %d\n", tp, fp, tn, fn);
		return Stats(tp,fp,tn,fn);
	}
};

#endif /* DATA_H_ */
