//BSD License - please cite paper for good karma :-)
/*
Copyright 2011 David Koes. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of David Koes.
*/

/*
 * main.cpp
 *
 *  Created on: Apr 25, 2011
 *      Author: dkoes
 *
 *      multIG - Multivariate Information Gain calculator.
 *      Computes multivariate information gain (ie, evaluates k atributes
 *      at a time and computes information of simultaneously splitting on
 *      all k).
 */

#include "CommandLine2/CommandLine.h"
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include "data.h"
#include "mvig.h"

using namespace std;

cl::opt<string> DataFile("data", cl::desc("Data file. First line must be header. Nonnumerical data is ignored."),cl::Required);
cl::opt<string> TestFile("test", cl::desc("Test file. Same format as data."));
cl::opt<bool> TestAll("test-all", cl::desc("Evaluate all computed rules on test file."), cl::init(false));
cl::opt<bool> BiasFalse("bias-false",cl::desc("Select cutoff directions that select for false examples."));
cl::opt<unsigned> K("k", cl::desc("number of attributes to evaluate"),cl::init(1));
cl::opt<unsigned> TopK("topk", cl::desc("compute joint information gain of k attributes with highest individual IG"), cl::init(0));
cl::opt<bool> Verbose("verbose", cl::desc("Verbose output"));
cl::opt<unsigned> Debug("debug-level", cl::desc("Debug output level."));
cl::opt<bool> Exhaustive("exhaustive",cl::desc("Consider all k combinations of attributes instead of greedily selecting (very very slow)"), cl::init(false));
cl::opt<string> Single("single",cl::desc("Compute information gain of single attribute."));
cl::list<string> Attributes("attribs", cl::desc("Compute information gain for specified set of attributes"));
cl::opt<unsigned> Samples("samples", cl::desc("Number of thresholds to sample uniformly (zero=all)"),cl::init(0));
cl::opt<unsigned> NThreads("nthreads", cl::desc("Number of threads to use"),cl::init(thread::hardware_concurrency()));
cl::opt<unsigned> CMax("cmax", cl::desc("Use cmax algorithm to find k best attributes"), cl::init(0));
cl::list<string> ReverseAttributes("reva", cl::desc("Force this attribute to use the suboptimal direction when splitting."));
//eventually want..
//greedy (vs optimal)
//target true/false classification
//best k vs. all k

int main(int argc, char *argv[])
{
	cl::ParseCommandLineOptions(argc, argv);

	ifstream infile(DataFile.c_str());
	if(!infile)
	{
		cl::PrintHelpMessage();
		cerr << "Could not open " << DataFile << "\n";
		exit(-1);
	}

	//read header
	string header;
	getline(infile, header);
	DataSchema schema(header);
	if(!schema.isValid())
	{
		cl::PrintHelpMessage();
		cerr << "Invalid header\n";
		exit(-1);
	}

	Data data(schema, infile);
	DataView alldata(data);

	if(BiasFalse)
		alldata.invert();

	for(unsigned i = 0, n = ReverseAttributes.size(); i < n; i++)
	{
		schema.setReverse(ReverseAttributes[i]);
	}

	MVIG mvig(schema, alldata);

	if(Single.length() > 0)
	{
		int index = schema.index(Single);
		if(index < 0)
		{
			cerr << "Invalid attribute.\n";
			exit(-1);
		}
		double split;
		double ig = mvig.computeInfoGain(index, split);
		printf("%s: %f at %f\n", Single.c_str(), ig, split);
	}
	else
	{
		vector<MVIGInfo> result;

		if(K > schema.size())
		{
			cerr << "K must be less than number of attributes.\n";
			exit(-1);
		}

		if(Attributes.size() > 0)
			mvig.computeAttributeInfoGains(Attributes, result);
		else if(CMax > 0)
			mvig.computeCMaxInfoGains(CMax, result);
		else if(TopK > 0)
			mvig.computeTopKInfoGains(TopK, result);
		else if(K == 1)
			mvig.computeAllSingleInfoGains(result);
		else if(Exhaustive)
			mvig.computeAllInfoGains(K, result);
		else
			mvig.computeGreedyInfoGains(K, result);

		sort(result.begin(), result.end());
		if (TestFile.size() == 0)
		{ //print out rules
			for (unsigned i = 0, n = result.size(); i < n; i++)
			{
				mvig.print(cout, result[i]);
				cout << "\n";
			}
		}
		else if (TestFile.size() > 0)
		{
			//print out rules and application results
			ifstream testfile(TestFile.c_str());
			if(!testfile)
			{
				cl::PrintHelpMessage();
				cerr << "Could not open " << TestFile << "\n";
				exit(-1);
			}

			//read header
			string header;
			getline(testfile, header);
			//assume same schema as data

			Data testdata(schema, testfile);
			DataView test(testdata);

			unsigned start = result.size()-1;
			if(TestAll) start = 0;
			for(unsigned i = start, n = result.size(); i < n; i++)
			{
				Stats stats = test.applyRule(result[i].rule, BiasFalse);
				mvig.print(cout, result[i]);
				cout << " | ";
				printf("[%.0f,%.0f,%.0f,%.0f] Accuracy %f  Precision %f  Recall %f  Specificity %f  F1 %f\n",
						stats.tp, stats.fp, stats.tn, stats.fn,
					stats.accuracy(), stats.precision(), stats.recall(), stats.specificity(), stats.F1());
			}
		}
	}

	return 0;
}
