//BSD License - please cite paper for good karma :-)
/*
Copyright 2011 David Koes. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of David Koes.
*/

/*
 * mvig.cpp
 *
 *  Created on: Apr 25, 2011
 *      Author: dkoes
 *
 *      Implementation of MVIG computation.
 */

#include "mvig.h"
#include <cmath>
#include <cstdio>

cl::opt<bool>
		ThreshSearch(
				"threshold-search",
				cl::desc(
						"Use efficient search (instead of brute force enumeration) for optimal threshold finding"),
				cl::init(true));
cl::opt<unsigned>
		VolumeCutoff(
				"vol-cutoff",
				cl::desc(
						"Volume to cutoff recursive search and switch to brute force"),
				cl::init(0));
cl::opt<unsigned> CacheSize("cache-size",
		cl::desc("Size of computed value cache"), cl::init(2000000));

enum ScoreType
{
	InformationGain, Accuracy, F1,Precision,Specificity
};

cl::opt<ScoreType>
		Scoring(
				cl::desc("Score function:"),
				cl::values(
						clEnumVal(InformationGain , "Shannon information gain"),



    clEnumVal(Accuracy, "Accuracy (tp+tn)/size")																		,
    clEnumVal(F1, "F1. Harmonic mean of recall and precision"),
    clEnumVal(Precision, "Precision tp/(tp+fp)"),
    clEnumVal(Specificity, "True Negative Rate tn/(tn+fp)"),
   clEnumValEnd) , cl::init(InformationGain));

//compute info gain of a single attribute and return the split point
double MVIG::computeInfoGain(unsigned index, double& split)
{
	if (data.size() < 1)
		return 0;
	split = singleSplits[index].split;
	return singleIG[index];
}

//generic scoring function, higher is better
double MVIG::score(unsigned selT, unsigned selF) const
{
	unsigned unselT = data.truesize() - selT;
	unsigned unselF = data.falsesize() - selF;
	double size = data.size();
	switch (Scoring)
	{
	case InformationGain:
	{
		if (selT + selF == 0)
			return 0;
		if (unselT + unselF == 0)
			return 0;

		double selSize = selT + selF;
		double unselSize = unselT + unselF;

		double selfratio = selF / selSize;
		double seltratio = selT / selSize;

		double unselfratio = unselF / unselSize;
		double unseltratio = unselT / unselSize;

		double selentropy = 0;
		if (selfratio != 0 && seltratio != 0)
			selentropy = -seltratio * log2(seltratio) - selfratio * log2(
					selfratio);
		double unselentropy = 0;
		if (unselfratio != 0 && unseltratio != 0)
			unselentropy = -unseltratio * log2(unseltratio) - unselfratio
					* log2(unselfratio);

		double entropy = selSize * selentropy / size + unselSize * unselentropy
				/ size;

		return H - entropy;
	}
	case Accuracy:
		return (selT + unselF) / size;
	case F1:
	{
		if(selT+unselT == 0)
			return 0;
		if(selT+selF == 0)
			return 0;
		double recall = selT/(double)(selT+unselT);
		double precision = selT/(double)(selT+selF);

		if(precision+recall == 0)
			return 0;
		return 2*precision*recall/(precision+recall);
	}
	case Precision:
		if(selT+selF == 0)
			return 0;
		return selT/(double)(selT+selF);
	case Specificity:
		if(unselF+selF == 0) return 0;
		return unselF/(double)(unselF+selF);
	}
	return 0;
}

double MVIG::computeScore(unsigned index, DataView::iterator& maxitr, bool reverse)
{
	if (data.size() < 1)
		return 0;

	//then evaluate conditional entropy for every possible split point
	//choosing split that maximizes the information gain
	data.sort(index);

	double max = 0;
	DataView::iterator itr = data.begin(index);
	maxitr = itr;
	for (unsigned i = 0, n = splits.all[index].size(); i < n; i++)
	{
		double split = splits.all[index][i];
		while (itr && itr.value() < split)
			itr.nextValue();

		double ig;
		if(!reverse)
			ig = score(itr.seenTrueCnt(), itr.seenFalseCnt());
		else
			ig = score(data.truesize()-itr.seenTrueCnt(), data.falsesize()-itr.seenFalseCnt());

		if (ig > max)
		{
			max = ig;
			maxitr = itr;
		}
	}

	return max;
}

//compute single attribute information gains
void MVIG::computeAllSingleInfoGains()
{
	for (unsigned index = 0, n = schema.size(); index < n; index++)
	{
		DataView::iterator pos;
		double ig = computeScore(index, pos, false);

		Split split;
		split.index = index;

		switch (Scoring)
		{
		case InformationGain:
		{
			//this isn't directional, just just see which way the cookie crumbles
			singleIG.push_back(ig);
			split.split = pos.value();

			//choose the direction that selects the hightest percent of true values
			double seentperc = pos.seenTrueCnt() / (double) pos.seenCnt();
			double remaintperc = (data.truesize() - pos.seenTrueCnt())
					/ (double) (data.size() - pos.seenCnt());
			split.greater = seentperc < remaintperc;

			singleSplits.push_back(split);

			//splits should be ordered so higher is the selected region
			if (!split.greater)
				splits.reverseDirection(index);
			break;
		}
		case Accuracy:
		case F1:
		case Precision:
		case Specificity:
			//these are directional
		{
			DataView::iterator rpos;
			double igr = computeScore(index, rpos, true);
			if (ig > igr)
			{
				//not reverse
				singleIG.push_back(ig);
				split.split = pos.value();
				split.greater = false;
				singleSplits.push_back(split);
				splits.reverseDirection(index);
			}
			else
			{
				singleIG.push_back(igr);
				split.split = rpos.value();
				split.greater = true;
				singleSplits.push_back(split);
			}
		}
		}

		if(schema.getReverse(index))
		{
			splits.reverseDirection(index);
			singleSplits.back().split = !singleSplits.back().split;
		}

	}
}

//put single infos in standard result form; informative only
void MVIG::computeAllSingleInfoGains(vector<MVIGInfo>& result)
{
	result.clear();
	result.reserve(singleIG.size());

	for (unsigned i = 0, n = singleIG.size(); i < n; i++)
	{
		if (singleIG[i] > 0)
		{
			MVIGInfo info;
			info.IG = singleIG[i];
			info.rule.push_back(singleSplits[i]);
			result.push_back(info);
		}
	}
}

//joint IG of columns a and b represented by their single rule
//memoized
double MVIG::jointIG(unsigned aindex, unsigned bindex)
{
	if (jointIGs[aindex][bindex] >= 0)
		return jointIGs[aindex][bindex];

	vector<unsigned> perm(2);
	perm[0] = aindex;
	perm[1] = bindex;

	MVIGInfo res;
	computeKInfoGain(perm, res);

	double IG = res.IG;
	jointIGs[aindex][bindex] = IG;
	jointIGs[bindex][aindex] = IG;

	return IG;
}

void MVIG::checkBest(const ThresholdSplitInfo& info,
		const vector<unsigned>& splitindices, MVIGInfo& bestres)
{
	double IG = info.IG;
	if (IG > bestres.IG)
	{
		//update
		bestres.IG = IG;
		for (unsigned i = 0, n = splitindices.size(); i < n; i++)
		{
			bestres.rule[i].split
					= splits.all[bestres.rule[i].index][splitindices[i]];
		}
	}
}

//unpack a bit
static void boxFromBits(unsigned long cbits, unsigned k,
		vector<unsigned>& lower, vector<unsigned>& upper,
		const vector<unsigned>& minsplitindex,
		const vector<unsigned>& midsplitindex,
		const vector<unsigned>& maxsplitindex)
{
	for (unsigned i = 0; i < k; i++)
	{
		if (cbits & (1 << i))
		{
			//midpoint for this dimension
			lower[i] = midsplitindex[i];
			upper[i] = maxsplitindex[i];
		}
		else
		{
			//minpoint for this dimension
			lower[i] = minsplitindex[i];
			upper[i] = midsplitindex[i];
		}
	}
}

struct IGCBits
{
	double IG;
	unsigned long cbits;

	IGCBits() :
		IG(0), cbits(0)
	{
	}
	IGCBits(double ig, unsigned long bits) :
		IG(ig), cbits(bits)
	{
	}

	bool operator<(const IGCBits& rhs) const
	{
		return IG > rhs.IG;
	}
};

//compute all the information gains within the specified bounding box
//(specified by indices into splits.all)
//and store the best if it is better
//rule must be properly initialized, this will just put in the split thresholds
void MVIG::computeBoxKInfoGain(unsigned pos, vector<Split>& rule,
		const vector<unsigned>& minsplitindex,
		const vector<unsigned>& maxsplitindex, MVIGInfo& bestres)
{
	if (pos == rule.size())
	{
		//calculate
		double IG = computeRuleInfo(rule).IG;
		if (IG > bestres.IG)
		{
			bestres.rule = rule;
			bestres.IG = IG;
		}
	}
	else //build up split
	{
		unsigned index = rule[pos].index;
		for (unsigned i = minsplitindex[pos], end = maxsplitindex[pos]; i < end; i++)
		{
			rule[pos].split = splits.all[index][i];
			computeBoxKInfoGain(pos + 1, rule, minsplitindex, maxsplitindex,
					bestres);
		}
	}
}

//this is a more efficient computer of the best set of thresholds for a joint
//information gain
//min/maxsplitindex define the bounding box of the attributes in terms of
//indices into the split arrays
//tsi stores (and computes) threshold infos for indices into splits
//these are precomputed
//best is the best information gain found so far, it should be initialized with
//a default rule indicating all the relevant attributes
void MVIG::computeRecursiveSearchKInfoGain(
		const vector<unsigned>& minsplitindex,
		const vector<unsigned>& maxsplitindex, TSICache& tsi, MVIGInfo& bestres)
{
	//compute midpoint
	unsigned k = minsplitindex.size(); //dimensionality
	if (k >= sizeof(unsigned long) * 8)
	{
		cerr << "Way too many attributes. Sorry.\n";
		exit(-1);
	}
	vector<unsigned> midsplitindex(k);
	unsigned volume = 1;
	for (unsigned i = 0; i < k; i++)
	{
		midsplitindex[i] = (maxsplitindex[i] + minsplitindex[i]) / 2;
		volume *= (maxsplitindex[i] - minsplitindex[i]);
	}

	if (midsplitindex == minsplitindex)
		return; //there is no midpoint
	if (volume <= VolumeCutoff)
	{
		vector<Split> rul = bestres.rule;
		computeBoxKInfoGain(0, rul, minsplitindex, maxsplitindex, bestres);
	}

	ThresholdSplitInfo mid = tsi.get(midsplitindex);
	checkBest(mid, midsplitindex, bestres);

	//iterate over all sub-hypercubes and compute best possible IG for each cube
	//we define a hypercube by the lowest corner which is either a minpoint or a midpoint
	vector<unsigned> lower(k, 0);
	vector<unsigned> upper(k, 0);
	vector<IGCBits> cubes;
	cubes.reserve(1 << k);

	for (unsigned long cbits = 0, end = (1 << k); cbits < end; cbits++)
	{
		boxFromBits(cbits, k, lower, upper, minsplitindex, midsplitindex,
				maxsplitindex);

		ThresholdSplitInfo upinfo = tsi.get(upper);
		ThresholdSplitInfo lowinfo = tsi.get(lower);

		checkBest(upinfo, upper, bestres);
		checkBest(lowinfo, lower, bestres);

		//compute best achievable IG and save
		//assuming thresholds are properly directional, all all true in region to selected and all falst to unselected
		if (lowinfo.falsecnt < upinfo.falsecnt)
		{
			cerr << "WARNING: splits are in a bad direction.\n";
		}

		double maxIG = score(lowinfo.truecnt, upinfo.falsecnt);
		assert(maxIG >= 0);
		cubes.push_back(IGCBits(maxIG, cbits));
	}

	//recurse on best cubes, skipping any where the best possible is worse than the current best
	sort(cubes.begin(), cubes.end());
	for (unsigned i = 0, n = cubes.size(); i < n; i++)
	{
		if (cubes[i].IG > bestres.IG) //need to explore
		{
			boxFromBits(cubes[i].cbits, k, lower, upper, minsplitindex,
					midsplitindex, maxsplitindex);
			computeRecursiveSearchKInfoGain(lower, upper, tsi, bestres);
		}
		else
		{
			break;
		}
	}

	//definitely no one will want this
	tsi.remove(midsplitindex);
}

//find the best thresholds for the k attribs and put the result in bestres
void MVIG::computeKInfoGain(const vector<unsigned>& attribs, MVIGInfo& bestres)
{
	if (ThreshSearch)
	{
		if (bestres.IG == 0) //not pre-initialized with good split, set it up
		{
			unsigned k = attribs.size();
			vector<Split> rule;
			rule.reserve(k);
			double max = 0;
			unsigned besti = 0;
			unsigned bestindex = 0;
			for (unsigned i = 0; i < k; i++)
			{
				int index = attribs[i];
				if(index < 0)
				{
					cerr << "Invalid attribute\n";
					exit(-1);
				}
				rule.push_back(singleSplits[index]);
				rule.back().split = splits.all[index].front(); //select all
				if (singleIG[index] > max)
				{
					max = singleIG[index];
					besti = i;
					bestindex = index;
				}
			}
			//initialize best to splitting on
			rule[besti].split = singleSplits[bestindex].split;
			bestres.rule = rule;
			bestres.IG = singleIG[bestindex];
		}

		unsigned k = bestres.rule.size();
		TSICache tsi(*this, bestres.rule);
		vector<unsigned> minsplitindex(k);
		vector<unsigned> maxsplitindex(k);
		for (unsigned i = 0; i < k; i++)
		{
			minsplitindex[i] = 0;
			maxsplitindex[i] = splits.all[attribs[i]].size() - 1;
		}

		computeRecursiveSearchKInfoGain(minsplitindex, maxsplitindex, tsi,
				bestres);

	}
	else //exhaustive enumeration
	{
		vector<Split> currule;
		computeBruteForceKInfoGain(0, attribs, currule, bestres);
	}
}

ThresholdSplitInfo MVIG::computeRuleInfo(vector<Split>& currule)
{
	//compute entropy of this rule
	AndSplitter splitter(currule);
	DataView A, B;
	data.splitter(splitter, A, B);

	ThresholdSplitInfo ret;
	ret.IG = score(A.truesize(), A.falsesize());
	ret.truecnt = A.truesize();
	ret.falsecnt = A.falsesize();

	return ret;
}

//compute the information gain of the join attribute specified by perm
//which indexes into singles
//works recursively, creating all possible rules for the p attribute
//and splitting
void MVIG::computeBruteForceKInfoGain(unsigned p, const vector<unsigned>& perm,
		vector<Split>& currule, MVIGInfo& bestres)
{
	if (p == perm.size())
	{
		//compute information gain of this rule
		double IG = computeRuleInfo(currule).IG;

		if (Debug > 2)
		{
			cout << "KIG ";
			for (unsigned i = 0, n = currule.size(); i < n; i++)
			{
				cout << currule[i].split << " ";
			}
			cout << IG << "\n";
		}

		if (IG > bestres.IG)
		{
			bestres.rule = currule;
			bestres.IG = IG;
		}
	}
	else
	{
		//get an initial split (will have right direction, change split value in loop)
		unsigned index = perm[p];
		Split split = singleSplits[index];
		currule.push_back(split);

		for (unsigned i = 0, n = splits.all[index].size(); i < n; i++)
		{
			currule.back().split = splits.all[index][i];
			computeBruteForceKInfoGain(p + 1, perm, currule, bestres);
		}
		currule.pop_back();
	}
}

//recursively generate permutations of informative attributes into result
static void genPerm(unsigned i, unsigned k, vector<unsigned>& curr,
		const vector<double>& singleIG, vector<vector<unsigned> >& result)
{
	if (curr.size() == k)
		result.push_back(curr);
	else
	{
		unsigned start = 0;
		if (curr.size() > 0)
			start = curr.back() + 1;
		for (unsigned pos = start, n = singleIG.size(); pos < n; pos++)
		{
			if (singleIG[pos] == 0)
				continue;
			curr.push_back(pos);
			genPerm(i + 1, k, curr, singleIG, result);
			curr.pop_back();
		}
	}
}

//considers all n choose k selections of attributes
void MVIG::computeAllInfoGains(unsigned k, vector<MVIGInfo>& result)
{
	//generate a n choose k permutations
	vector<vector<unsigned> > permutations;
	vector<unsigned> curr;
	genPerm(0, k, curr, singleIG, permutations);

	if (permutations.size() == 0)
	{
		cerr << "K must be less than number of informative attributes.\n";
		exit(-1);
	}

	thread_group group;
	result.resize(permutations.size());
	for (unsigned i = 0; i < NThreads; i++)
	{
		KInfoThread th(*this, permutations, i, NThreads, result);
		group.add_thread(new thread(th));
	}

	group.join_all();

}

//computes best IG for specified attributes
void MVIG::computeAttributeInfoGains(const vector<string>& attribs,
		vector<MVIGInfo>& result)
{

	vector<unsigned> indices;
	//get index  of specified attributes
	for (unsigned i = 0, n = attribs.size(); i < n; i++)
	{
		int index = schema.index(attribs[i]);
		if(index < 0)
		{
			cerr << "Invalid attribute " << attribs[i] << "\n";
			exit(-1);
		}
		indices.push_back(index);
	}

	MVIGInfo res;
	computeKInfoGain(indices, res);

	result.push_back(res);

}

void MVIG::printInfo(const MVIGInfo& res)
{
	cout << res.IG;
	for (unsigned j = 0, m = res.rule.size(); j < m; j++)
	{
		unsigned index = res.rule[j].index;
		cout << "  " << schema.name(index) << " (";
		if (res.rule[j].greater)
			cout << ">=";
		else
			cout << "<";
		cout << res.rule[j].split << ")";
	}
	cout << endl;
}

//greedily selects attributes, but fully optimizes thresholds of selected attributes
//result has the incrimental rules as well
void MVIG::computeGreedyInfoGains(unsigned k, vector<MVIGInfo>& result)
{
	vector<bool> isselected(singleIG.size(), false);
	vector<unsigned> selected;

	//choose highest information gain
	double max = 0;
	unsigned best = 0;
	for (unsigned i = 0, n = singleIG.size(); i < n; i++)
	{
		if (singleIG[i] > max)
		{
			max = singleIG[i];
			best = i;
		}
	}

	selected.push_back(best);
	isselected[best] = true;

	result.push_back(MVIGInfo(max, singleSplits[best]));

	for (--k; k > 0; k--)
	{
		MVIGInfo bestres;
		//try every possible unselected attribute
		best = 0;
		for (unsigned i = 0, n = singleIG.size(); i < n; i++)
		{
			if (!isselected[i])
			{
				selected.push_back(i);
				vector<Split> currule;
				//initialize rule to best so far
				MVIGInfo res = result.back();
				res.rule.push_back(singleSplits[i]);
				res.rule.back().split = splits.all[i].back();
				computeKInfoGain(selected, res);
				selected.pop_back();
				if (res.IG > bestres.IG)
				{
					bestres = res;
					best = i;
				}
			}
		}
		isselected[best] = true;
		selected.push_back(best);
		result.push_back(bestres);

		if (Verbose)
		{
			printInfo(result.back());
		}
	}
}

//compute joint information gain for topk best ranked attributes
void MVIG::computeTopKInfoGains(unsigned topk, vector<MVIGInfo>& result)
{

	vector<unsigned> top;
	top.reserve(topk);
	vector<bool> isselected(singleIG.size(), false);
	for (unsigned t = 0; t < topk; t++)
	{
		double max = 0;
		unsigned best = 0;
		for (unsigned i = 0, n = singleIG.size(); i < n; i++)
		{
			//simple insertion sort
			if (!isselected[i] && singleIG[i] > max)
			{
				max = singleIG[i];
				best = i;
			}
		}
		top.push_back(best);
		isselected[best] = true;
	}

	vector<Split> currule;
	MVIGInfo res;
	computeKInfoGain(top, res);
	result.push_back(res);
}

struct IndexIG
{
	double IG;
	unsigned index;

	IndexIG(unsigned i, double ig) :
		IG(ig), index(i)
	{
	}

	bool operator<(const IndexIG& rhs) const
	{
		return IG > rhs.IG;
	}
};

//uses "cmax" algorithm to select attributes - first compute all pairwise
//information gains, then choose the best pair and iterative select a new attribute
//until we have k attributes
//we select the attribute with the maximum minimum information gain relative to already selected attributes
//then we reoptimize
void MVIG::computeCMaxInfoGains(unsigned k, vector<MVIGInfo>& result)
{

	//find best initial attribute
	unsigned best;
	double max = 0;
	for (unsigned i = 0, n = singleIG.size(); i < n; i++)
	{
		double IG = singleIG[i];
		if (IG > max)
		{
			max = IG;
			best = i;
		}
	}

	//memoize pairwise information
	MVIGInfo first(max, singleSplits[best]);
	result.push_back(first);

	if (Verbose)
		printInfo(result[0]);

	//now add additional features

	vector<bool> isselected(singleIG.size(), false);
	isselected[best] = true;
	vector<unsigned> selected;
	selected.push_back(best);

	for (unsigned a = 1; a < k; a++)
	{

		//multithread computation of new joint igs
		thread_group group;
		for (unsigned i = 0; i < NThreads; i++)
		{
			JointInfoThread th(*this, i, NThreads, selected.back());
			group.add_thread(new thread(th));
		}
		group.join_all();

		vector<IndexIG> choices;
		for (unsigned i = 0, n = singleIG.size(); i < n; i++)
		{
			if (isselected[i])
				continue;

			double min = 1;
			for (unsigned j = 0, m = selected.size(); j < m; j++)
			{
				double condIG = jointIG(i, selected[j]) - singleIG[selected[j]];
				if (condIG < min)
				{
					min = condIG;
				}
			}

			choices.push_back(IndexIG(i, min));

		}

		//select the attribute with the maximum conditional mutual information
		sort(choices.begin(), choices.end());

		unsigned i, n;
		for (i = 0, n = choices.size(); i < n; i++)
		{
			unsigned choice = choices[i].index;
			//create an initial rule from previous result
			MVIGInfo res(result.back());
			//add new attribute with no selectivity
			res.rule.push_back(singleSplits[choice]);
			res.rule.back().split = splits.all[choice].back();
			selected.push_back(choice);

			computeKInfoGain(selected, res);

			//check for improvement
			if (res.IG > result.back().IG)
			{
				isselected[choice] = true;
				result.push_back(res);

				if (Verbose)
					printInfo(res);
				break;
			}
			else
			{
				//keep looking
				selected.pop_back();
				if (Verbose)
					cout << schema.name(choice) << " uninformative\n";
			}
		}
		if (i == n)
			break; //nothing good
	}

}

//print out ig and rule
void MVIG::print(ostream& out, const MVIGInfo& info) const
{
	out << info.IG;
	for (unsigned j = 0, m = info.rule.size(); j < m; j++)
	{
		unsigned index = info.rule[j].index;
		out << "  " << schema.name(index) << " (";
		if (info.rule[j].greater)
			out << ">=";
		else
			out << "<";
		out << info.rule[j].split << ")";
	}
}

