//BSD License - please cite paper for good karma :-)
/*
Copyright 2011 David Koes. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of David Koes.
*/

/*
 * mvig.h
 *
 *  Created on: Apr 25, 2011
 *      Author: dkoes
 */

#ifndef MVIG_H_
#define MVIG_H_

#include "data.h"
#include <boost/thread.hpp>

extern cl::opt<unsigned> Samples;
extern cl::opt<unsigned> NThreads;

extern cl::opt<unsigned> CacheSize;


//complete information for the threshold split
struct ThresholdSplitInfo
{
	unsigned truecnt; //selected by split -> unselected is just remainder
	unsigned falsecnt;
	double IG;

	ThresholdSplitInfo(): truecnt(0), falsecnt(0), IG(-1) {}

	ThresholdSplitInfo(unsigned tc, unsigned fc, double e): truecnt(tc), falsecnt(fc), IG(e)
	{

	}
};



//a list of all unique datavalues for every attribute in a dataset
struct AllSplits
{
	vector< vector<double> > all;

	AllSplits() {}
	//sorts data
	AllSplits(DataView& data)
	{
		all.resize(data.columns());
		for(unsigned i = 0, n = data.columns(); i < n; i++)
		{
			data.sort(i);
			//sentinal values at begining..
			all[i].push_back(-HUGE_VAL);

			if(Samples > 0)
			{
				double step = data.size()/(double)Samples;
				if(step < 1) step = 1.0;
				unsigned pos = 0;
				double val = 0;
				for(DataView::iterator itr = data.begin(i); itr; ++itr,pos++)
				{
					if(pos == (unsigned)val)
					{
						all[i].push_back(itr.value());
						val += step;
					}
				}

				vector<double>::iterator last = unique(all[i].begin(), all[i].end());
				all[i].erase(last, all[i].end());
			}
			else //all values
			{
				for(DataView::iterator itr = data.begin(i); itr; itr.nextValue())
				{
					all[i].push_back(itr.value());
				}
			}

			//end value that's larger than everything
			all[i].push_back(HUGE_VAL);
		}
	}

	//flip the ordering of values so a higher index is always more selective
	void reverseDirection(unsigned index)
	{
		reverse(all[index].begin(), all[index].end());
	}
};




//associates a spliting rule with information gain
struct MVIGInfo
{
	vector<Split> rule; //a partition
	double IG; //information gain

	MVIGInfo(): IG(0) {}

	MVIGInfo(double ig, const Split& single): IG(ig)
	{
		rule.push_back(single);
	}

	bool operator<(const MVIGInfo& rhs) const
	{
		return IG < rhs.IG;
	}

};

class TSICache;
/* compute multi-variate information gain in different ways */
class MVIG
{
	DataSchema schema;
	DataView data;
	AllSplits splits;
	vector<Split> singleSplits;  //indexed by index
	vector<double> singleIG; //indexed by index
	vector< vector<double> > jointIGs;

	double H;

	double computeScore(unsigned index, DataView::iterator& minitr, bool reverse);
	void computeBruteForceKInfoGain(unsigned p, const vector<unsigned>& perm,
			vector<Split>& currule, MVIGInfo& bestres);

	void printInfo(const MVIGInfo& res);

	//memoized jointIG
	double jointIG(unsigned aindex, unsigned bindex);

	ThresholdSplitInfo computeRuleInfo(vector<Split>& currule);

	void checkBest(const ThresholdSplitInfo& info, const vector<unsigned>& splitindices, MVIGInfo& bestres);
	void computeBoxKInfoGain(unsigned pos, vector<Split>& rule, const vector<unsigned>& minsplitindex,
			const vector<unsigned>& maxsplitindex,
			MVIGInfo& bestres);
	void computeRecursiveSearchKInfoGain(const vector<unsigned>& minsplitindex,
			const vector<unsigned>& maxsplitindex,
			TSICache& tsi,
			MVIGInfo& bestres);

	void computeKInfoGain(const vector<unsigned>& attribs, MVIGInfo& bestres);

	//compute info gains (with directions for filter) for all individual attributes
	void computeAllSingleInfoGains();

	friend class KInfoThread;
	friend class JointInfoThread;
	friend class TSICache;
public:

	/* Since we modify (sort) the data store it as a copy */
	MVIG(const DataSchema& sch, const DataView& dat): schema(sch), data(dat), splits(data),
	jointIGs(data.columns(), vector<double>(data.columns(), -1))
	{
		H = data.entropy();
		computeAllSingleInfoGains();
	}

	//compute the information gain of a specified attribute
	//returns the best information gain and the split point
	double computeInfoGain(unsigned index, double& split);

	//compute info gains (with directions for filter) for all individual attributes
	void computeAllSingleInfoGains(vector<MVIGInfo>& result);

	//considers all n choose k selections of attributes
	void computeAllInfoGains(unsigned k, vector<MVIGInfo>& result);

	//greedily selects attributes for inclusion, but optimizes thresholds
	void computeGreedyInfoGains(unsigned k, vector<MVIGInfo>& result);

	//computes best IG for specified attributes
	void computeAttributeInfoGains(const vector<string>& attribs, vector<MVIGInfo>& result);

	void computeTopKInfoGains(unsigned topk, vector<MVIGInfo>& result);

	void computeCMaxInfoGains(unsigned k, vector<MVIGInfo>& result);

	void print(ostream& out, const MVIGInfo& info) const;

	double score(unsigned selT, unsigned selF) const;
};


//stores ThresholdSplitInfo indexed by a vector of unsigned
//which are indices into all splits in attribute order
//in theory it is possible to figure out when keys become uninteresting
//and remove them manually in the caller, but for simplicity just implement a cache
class TSICache
{
	MVIG& mvig;
	typedef pair<ThresholdSplitInfo, const vector<unsigned>* > CacheType;
	list<CacheType> cache;
	typedef unordered_map< vector<unsigned>, list< CacheType >::iterator > IndexType;
	IndexType index;
	vector<Split> protorule; //sample rule, just thresholds change
	unsigned cachesize; //list size is linear

public:
	TSICache(MVIG& mv, const vector<Split>& r): mvig(mv), protorule(r), cachesize(0)
	{
	}

	ThresholdSplitInfo get(const vector<unsigned>& key)
	{
		IndexType::iterator pos = index.find(key);
		if(pos != index.end())
		{
			list<CacheType>::iterator itr = pos->second;
			ThresholdSplitInfo info = itr->first;
			//move to front (LRU)
			cache.push_front(*itr);
			pos->second = cache.begin();
			cache.erase(itr);

			return info;
		}
		else
		{
			//compute and insert
			//create rule
			assert(key.size() == protorule.size());
			for(unsigned i = 0, n = protorule.size(); i < n; i++)
			{
				unsigned index = protorule[i].index;
				double thresh = mvig.splits.all[index][key[i]];
				protorule[i].split = thresh;
			}
			ThresholdSplitInfo info = mvig.computeRuleInfo(protorule);
			cache.push_front(CacheType(info, NULL));
			cachesize++;
 			index[key] = cache.begin();
 			//get pointer to key within hash
 			IndexType::iterator pos = index.find(key);
 			cache.begin()->second = &pos->first;

 			while(cachesize > CacheSize)
 			{
 				//erase from hash
 				index.erase(*cache.back().second);
 				//remove from list
 				cache.pop_back();
 				cachesize--;
 			}
			return info;
		}
	}

	//if we know we'll never need a value again, save memory and remove
	void remove(const vector<unsigned>& key)
	{
		IndexType::iterator pos = index.find(key);
		if(pos != index.end())
		{
			cache.erase(pos->second);
			cachesize--;
			index.erase(key);
		}
	}
};

//wrapper for running a bunch of kinfo gains
class KInfoThread
{
	MVIG& mvig;
	const vector<vector<unsigned> >& perms;
	vector<MVIGInfo>& results;
	unsigned start;
	unsigned stride;
public:
	//res must be properly sized
	KInfoThread(MVIG& mv, const vector<vector<unsigned> >& ps, unsigned st, unsigned str,
			vector<MVIGInfo>& res): mvig(mv),
				perms(ps), results(res), start(st), stride(str)
	{
		assert(results.size() == perms.size());
	}

	void operator()()
	{
		for(unsigned i = start, n = perms.size(); i < n; i+= stride)
		{
			mvig.computeKInfoGain(perms[i], results[i]);
		}
	}
};

//wrapper from computing a bunch of joint infos with respect to a single provided attribute
//results get memoized
class JointInfoThread
{
	MVIG& mvig;
	unsigned start;
	unsigned stride;
	unsigned sindex;

public:
	//res must be properly sized
	JointInfoThread(MVIG& mv, unsigned st, unsigned str,
			unsigned si): mvig(mv),
				start(st), stride(str), sindex(si)
	{
	}

	void operator()()
	{
		//find our slice of joints with sindex
		for(unsigned i = start, n = mvig.singleIG.size(); i < n; i+= stride)
		{
			if(i != sindex)
				mvig.jointIG(sindex, i);
		}
	}
};
#endif /* MVIG_H_ */
